@@ -50,8 +50,10 @@ class TestCase(unittest.TestCase):
                 zip.writestr(file_name, self.get_data(res_name))
             zip.writestr('pack.txt', '{0}\n{1}'.format(name, desc).encode('UTF-8'))
 
-    def assertRepresentIdenticalImages(self, bytes1, bytes2, msg=None):
+    def assert_PNGs_match(self, bytes1, bytes2, msg=None):
         im1 = Image.open(StringIO(bytes1))
+        if isinstance(bytes2, ResourceBase):
+            bytes2 = bytes2.get_bytes()
         im2 = Image.open(StringIO(bytes2))
         self.assertEqual(im1.size, im2.size)
         w, h = im1.size
@@ -291,7 +293,7 @@ class CompositeResourceTests(TestCase):
 
         # Check it matches the manually created image.
         bytes = res.get_bytes()
-        self.assertRepresentIdenticalImages(self.get_data('a_b_replace.png'), bytes)
+        self.assert_PNGs_match(self.get_data('a_b_replace.png'), bytes)
 
 
 class ExternalResourceTests(TestCase):
@@ -333,6 +335,10 @@ class ExternalResourceTests(TestCase):
         spec = self.loader.maybe_get_spec({'file': self.file_path}, base=None)
         self.assertEqual(self.stuff, spec)
 
+    def test_get_spec_infer_suffix(self):
+        spec = self.loader.maybe_get_spec({'file': os.path.join(self.test_dir, 'stuff')}, base=None, ext='json')
+        self.assertEqual(self.stuff, spec)
+
     def test_get_spec_yaml(self):
         file_path = os.path.join(self.test_dir, 'nonsense.tpmaps')
         with open(file_path, 'wt') as strm:
@@ -379,7 +385,11 @@ class ExternalResourceTests(TestCase):
         self.assertEqual('hello', self.loader.get_bytes('bambi', 'bim:///gooshy/gooshy/gander'))
         self.assertEqual('///gooshy/gooshy/bambi', mock_func.call_args[0][0])
 
+
 class ResolveUrlTests(unittest.TestCase):
+    # I ended up creating my own generic URL resolver, because
+    # the standard library’s urljoin seems reluctant to tackle
+    # unknown URL schemes. (Not entirely unreasonably.)
     def test_relative(self):
         self.assertEqual('foop:///derp/yum', resolve_generic_url('foop:///derp/herp', 'yum'))
 
@@ -405,6 +415,39 @@ class ResolveUrlTests(unittest.TestCase):
         self.assertEqual('shump://bank/flank/gazank',
                 resolve_generic_url('foop://bink/fink', 'shump://bank/flank/gazank'))
 
+    def test_dotdot(self):
+        self.assertEqual('smuurf://smurf/derf/lerf',
+                resolve_generic_url('smuurf://smurf/derf/bink/bank.frankly.dank', '../lerf'))
+
+    def test_dotdotdotdot(self):
+        self.assertEqual('smuurf://smurf/lerf',
+                resolve_generic_url('smuurf://smurf/derf/bink/bank.frankly.dank', '../../lerf'))
+
+    def test_dotdot_respects_authority(self):
+        self.assertEqual('smuurf://smurf/lerf',
+                resolve_generic_url('smuurf://smurf/derf/bink/bank.frankly.dank', '../../../lerf'))
+
+    def test_dot(self):
+        self.assertEqual('smuurf://smurf/derf/bink/lerf',
+                resolve_generic_url('smuurf://smurf/derf/bink/bank.frankly.dank', './lerf'))
+
+    def test_dotslash(self):
+        self.assertEqual('smuurf://smurf/derf/bink/',
+                resolve_generic_url('smuurf://smurf/derf/bink/bank.frankly.dank', './'))
+
+    def test_nothing_at_all(self):
+        self.assertEqual('smuurf://smurf/derf/bink/bank.frankly.dank',
+                resolve_generic_url('smuurf://smurf/derf/bink/bank.frankly.dank', ''))
+
+    def test_something_dotdot_something_for_heavens_sake(self):
+        self.assertEqual('smuurf://smurf/derf/bink/bink/bib/nib/jib',
+                resolve_generic_url('smuurf://smurf/derf/bink/bink/frothwagon', 'bib/fib/../nib/jib'))
+
+    def test_something_dotdot_sneak_attacj(self):
+        self.assertEqual('smuurf://smurf/etc/passwd',
+                resolve_generic_url('smuurf://smurf/derf/bink/bink/frothwagon', 'bib/../../../../../../../../../../etc/passwd'))
+
+
 class MixerTests(TestCase):
     def test_get_pack_by_name(self):
         pack1 = self.sample_pack()
@@ -433,7 +476,7 @@ class MixerTests(TestCase):
                 self.assertEqual(pack1.get_resource(n).get_content(),
                     pack2.get_resource(n).get_content())
             else:
-                self.assertRepresentIdenticalImages(
+                self.assert_PNGs_match(
                         pack1.get_resource(n).get_bytes(),
                         pack2.get_resource(n).get_bytes())
 
@@ -509,6 +552,11 @@ class MixerTests(TestCase):
         }, data1)
 
         pack2 = Mixer().get_pack('http://example.org/frog.zip')
+        # Not laoded yet:
+        self.assertFalse(mock_meth.call_args)
+
+        res = pack2.get_resource('a.png')
+        # Has now downloaded the data:
         self.assertEqual('http://example.org/frog.zip', mock_meth.call_args[0][0])
         self.assert_same_packs(pack1, pack2)
 
@@ -538,9 +586,12 @@ class MixerTests(TestCase):
             'status': '404',
         }, 'Not found')
 
-        with self.assertRaises(UnknownPack):
+        with self.assertRaises(CouldNotLoad):
             pack2 = Mixer().get_pack({'href': 'http://example.org/frog.zip'})
 
+            # Force it to be loaded:
+            res = pack2.get_resource('a.png')
+
     def test_get_pack_from_minecraft(self):
         pack1, data1 = self.sample_pack_and_bytes()
 
@@ -1208,7 +1259,7 @@ class MixerTests(TestCase):
         strm.seek(0)
         with ZipFile(strm, 'r') as zip:
             for file_name, resource_name in expected_contents.items():
-                self.assertRepresentIdenticalImages(
+                self.assert_PNGs_match(
                         self.get_data(resource_name),
                         zip.read(file_name),
                         'Expected contents of {actual} to match {expected}'.format(
@@ -1248,6 +1299,11 @@ class TestLastModified(TestCase):
             'Expected {0} <= {1} <= {2} within {3}'.format(
                 before, actual, after, epsilon))
 
+    def check_modified_since(self, x):
+        self.assertTrue(x.is_modified_since(x.get_last_modified() + timedelta(seconds=-1)))
+        self.assertFalse(x.is_modified_since(x.get_last_modified()))
+        self.assertFalse(x.is_modified_since(x.get_last_modified() + timedelta(seconds=1)))
+
     def test_file_pack(self):
         simple_map = GridMap((32, 32), (16, 16), ['a', 'b', 'c', 'd'])
 
@@ -1274,6 +1330,13 @@ class TestLastModified(TestCase):
         dt = datetime(*fake_time)
         self.assert_datetime_between(dt, res_b.get_last_modified(), dt, timedelta(seconds=2))
 
+        self.check_modified_since(res_a)
+        self.check_modified_since(res_b)
+
+        for n in pack.get_resource_names():
+            print n, pack.get_resource(n).is_modified_since(pack.get_last_modified())
+        self.check_modified_since(pack)
+
     def test_directory_pack(self):
         dir_path = os.path.join(self.test_dir, 'happipakq')
         if os.path.exists(dir_path):
@@ -1284,7 +1347,7 @@ class TestLastModified(TestCase):
             strm.write(self.get_data('a.png'))
         after = datetime.now()
         shutil.copy2(os.path.join(self.data_dir, 'b.png'),
-                os.path.join(self.test_dir, 'b.png'))
+                os.path.join(dir_path, 'b.png'))
         # Since that copied the last-modified time of the file,
         # we now have a dir containing files with 2 different modified times.
 
@@ -1294,6 +1357,132 @@ class TestLastModified(TestCase):
         res_a = pack.get_resource('a.png')
         self.assert_datetime_between(before, res_a.get_last_modified(), after,timedelta(seconds=1))
 
+        res_b = pack.get_resource('b.png')
+        t = datetime.fromtimestamp(os.stat(os.path.join(dir_path, 'b.png')).st_mtime)
+        self.assert_datetime_between(t, res_b.get_last_modified(), t,timedelta(seconds=1))
+
+        self.check_modified_since(res_a)
+        self.check_modified_since(res_b)
+        self.check_modified_since(pack)
+
+    def test_composite_resource(self):
+        # Create 2 resources
+        simple_map = GridMap((32, 32), (16, 16), ['a', 'b', 'c', 'd'])
+
+        file_path = os.path.join(self.test_dir, 'aa.zip')
+        before = datetime.now()
+        with open(file_path, 'wb') as strm:
+            self.write_pack_contents(strm,'AB', 'Has A and B',
+                    {'a.png': ('a.png', simple_map)})
+        after = datetime.now()
+
+        fake_time = (2010, 2, 20, 10, 39, 55)
+        with ZipFile(file_path, 'a') as zip:
+            zip.writestr(ZipInfo('b.png', fake_time), self.get_data('b.png'))
+
+        pack = SourcePack(file_path, Atlas({'a.png': simple_map, 'b.png': simple_map}))
+        res = CompositeResource('ab.png', pack.get_resource('b.png'), simple_map)
+        res.replace(pack.get_resource('a.png'), simple_map, {'a': 'd', 'd': 'a'})
+        self.assert_datetime_between(before, res.get_last_modified(), after, timedelta(seconds=2))
+
+        self.check_modified_since(res)
+        self.check_modified_since(pack)
+
+    def test_renamed_resource(self):
+        before = datetime.now()
+        pack = self.make_source_pack('Sign pack', 'Just a test', {'item/sign.png': ('sign.png', None)})
+        after = datetime.now()
+        res = RenamedResource('portent.png', pack.get_resource('item/sign.png'))
+        self.assert_datetime_between(before, res.get_last_modified(), after, timedelta(seconds=2))
+
+        self.check_modified_since(res)
+
+
+class PackPngTests(TestCase):
+    def setUp(self):
+        self.map = GridMap((256, 256), (16, 16), ['{0:02X}'.format(x) for x in range(256)])
+        self.pack = self.make_source_pack('Zumpy', 'Ging', {'terrain.png': ('gingham.png', self.map)})
+
+    def test_10(self):
+        res = PackIconResource(self.pack.get_resource('terrain.png'), self.map,
+            ['ED', 'DA', 'EB', 'FE', '31', 'ED', 'AC', 'BF', 'DA', '96'])
+            # Was DEADBEEF13DECAFBAD69 but I got the coordinates reversed in the test image.
+        with open(os.path.join(self.test_dir, 'pack10.png'), 'wb') as strm:
+            strm.write(res.get_bytes())
+        self.assert_PNGs_match(self.get_data('gingham10.png'), res.get_bytes())
+
+    def test_recipe_10(self):
+        m = Mixer()
+        m.add_pack('gingham', self.pack)
+        pack = m.make({
+            'label': 'lab',
+            'desc': 'desc',
+            'mix': {
+                'pack': '$gingham',
+                'files': [
+                    'terrain.png',
+                    {
+                        'source': 'terrain.png',
+                        'pack_icon': {
+                            'cells': ['ED', 'DA', 'EB', 'FE', '31', 'ED', 'AC', 'BF', 'DA', '96'],
+                        }
+                    }
+                ]
+            }
+        })
+        print pack.get_resource_names()
+        with open(os.path.join(self.test_dir, 'recipe_10.zip'), 'wb') as strm:
+            pack.write_to(strm)
+        self.assert_PNGs_match(self.get_data('gingham10.png'),
+            pack.get_resource('pack.png').get_bytes())
+
+
+class GuessPackTests(TestCase):
+    def test_from_zip(self):
+        zip_path = os.path.join(self.test_dir, 'brukken.zip')
+        with ZipFile(zip_path, 'w') as zip:
+            zip.writestr('terrain.png', self.get_data('gingham.png'))
+            zip.writestr('sign.png', self.get_data('sign.png'))
+            zip.writestr('irrelevant.png', self.get_data('a.png'))
+
+        atlas = Atlas({
+            'terrain.png': GridMap((256, 256), (16, 16), ['{0:02X}'.format(x) for x in range(256)]),
+            'item/sign.png': None,
+        })
+        pack = Mixer().make({
+            'label': 'unjumbled',
+            'desc': 'untwisted',
+            'mix': {
+                'pack': {
+                    'href': url_from_file_path(zip_path),
+                    'unjumble': {
+                        'terrain.png': {
+                            'source_rect': {'width': 256, 'height': 256},
+                            'cell_rect': {'width': 16, 'height': 16},
+                            'names': ['{0:02X}'.format(x) for x in range(256)],
+                        },
+                        'item/sign.png': None,
+                    },
+                },
+                'files': [
+                    '*.png',
+                    {
+                        'source': 'terrain.png',
+                        'pack_icon': {
+                            'cells': ['ED', 'DA', 'EB', 'FE', '31', 'ED', 'AC', 'BF', 'DA', '96'],
+                        }
+                    }
+                ]
+            }
+        })
+        with open(os.path.join(self.test_dir, 'unjumbled.zip'), 'w') as strm:
+            pack.write_to(strm)
+        self.assert_PNGs_match(self.get_data('gingham.png'), pack.get_resource('terrain.png'))
+        self.assert_PNGs_match(self.get_data('sign.png'), pack.get_resource('item/sign.png'))
+        self.assert_PNGs_match(self.get_data('gingham10.png'), pack.get_resource('pack.png'))
+        self.assertEqual('unjumbled', pack.label)
+        self.assertEqual('untwisted', pack.desc)
+
 
 if __name__ == '__main__':
-	unittest.main()
\ No newline at end of file
+	unittest.main()
